#!/usr/bin/perl

##################################################################
# Copyright (C) 1998 Stefan Mashkevich <mash@mashke.org>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License.
# Disclaimer: The program is provided "as is" and there is no
# warranty of any kind whatsoever. Your using it is at your own risk.
##################################################################
# Transliterating from KOI8, DOS866, WIN1251, ISO8859-5, MAC
# into any of these or into Latin characters
# Manually: between two given encodings
# Automatically: into a given encoding, figuring out the initial one
# automatically by frequency analysis
#
# v0.1 S.Mashkevich  16 Nov 1998
#      first working version
# v0.5 S.Mashkevich  26 Nov 1998
#      redesigned and shortened main code
#      completely rewrote frequency analysis algorithm
#      added MAC, ISO, Ukrainian and Belorussian characters
# v0.8 S.Mashkevich  13 Dec 1998
#      added reverse transliteration (handled by separate script;
#      this script was not changed)
#
# text on standard input, recoded on standard output
# see usage message
#
##################################################################
# modified for Georgian encodings:
# BPG  GOV    Latin Translit
# v0.9 Levan shoshiashvili 23 Aug 2000

@intos = qw/w k l /;

$ncods = @intos -1 ;  #  last = Latin, treated separately
##################################################################
# read in character codes

while (<DATA>) {
    /^\d/ && do {
        @data = split;
        for $j (0..$ncods-1) {
            $chars{$intos[$j]} = join '', $chars{$intos[$j]}, chr($data[$j]);
            $latin{chr($data[$0])} = $data[$ncods];
            }
    }
}

##################################################################
# determine what to do

@args = split //, $ARGV[0]; 

# into what
$into = pop @args || usage();
# If Latin, will recode into 0 and raise flag to re-recode later on
$into eq $intos[$ncods] && do {$into = $intos[0]; $latflag = 1};
$chars_into = $chars{$into} || usage();

# from what (or automatic)
defined ($from = pop @args) ?
    ($chars_from = $chars{$from} || usage()) :
    ($automatic = 1) ;

# there shouldn't be anything else in the argument
defined (pop @args) && usage();

##################################################################
# actual recoding

undef $/;
$_ = <STDIN>;

# analyze if necessary
defined ($automatic) && do {
    $result = &analyze;
    do {print; exit} unless ($result);   # probably non-Russian text
    $chars_from = $chars{$result};
    };
    
# transliterate
eval "tr/$chars_from/$chars_into/, 1" || die $@;

# if it was Latin
defined ($latflag) && do {
    foreach $char (keys %latin) {
        s/$char/$latin{$char}/g;
        };
    };

# output
print;

##################################################################
sub analyze {

# Frequency analyzer of Russian texts
# attempting to determine which encoding they are in
# Returns 0 if $_ appears to be non-Russian text
# or one of k,a,w,m,i as it deems $_ is in
# Things really depend on specific features of the encodings here,
# therefore I don't care to avoid hardwired "magic constants"

# first, count how many (perhaps) Russian characters are there
my $intrst_chars = join '', map(chr, 128..255);
my $length = eval "tr /$intrst_chars/$intrst_chars/";

# threshold for deciding when there are "too few" frequent characters in it
my $thrhold = 0.08;

# letters we will use: a,e,i,o,A,E,I,O
my %aeio, %num_aeio, $enc_aieo, %AEIO, %num_AEIO, $enc_AIEO, $enc;
my $num_aeio_max = 0, $num_AIEO_max = 0;
$aeio{k} = join '', map(chr, qw/192 196 200 205/);
$aeio{w} = join '', map(chr, qw/192 196 201 207/);
$AEIO{k} = join '', map(chr, qw/192 196 200 205/);
$AEIO{w} = join '', map(chr, qw/192 196 201 207/);


# Count how many times there are "a, e, i, o"
for $enc (qw/k w /) {
    $num_aeio{$enc} = eval "tr/$aeio{$enc}/$aeio{$enc}/";
    $num_aeio{$enc} > $num_aeio_max  &&  do {
        $num_aeio_max = $num_aeio{$enc}; $enc_aeio = $enc
        }
    };

# suppose we found them (relative number more than $freqthr)
$num_aeio_max > $thrhold * $length  &&  do {
    # if 'a', no doubt
    $enc_aeio eq 'a'  &&  return 'a';
    # if 'k' or 'i', might really be 'w' capitals; check this
    ($enc_aeio eq 'k' || $enc_aeio eq 'i')  &&  do {
        $num_AEIO{w} = eval "tr/$AEIO{w}/$AEIO{w}/";
        $num_AEIO{w} > $num_aeio_max  ?  return 'w'  :  return $enc_aeio
        };
    # otherwise (i.e 'w') decide between 'w' and 'm' using caps
    my $wincaps = join '', map(chr, 192..255);
    my $maccaps = join '', map(chr, 128..159);
    my $num_wincaps = eval "tr/$wincaps/$wincaps/";
    my $num_maccaps = eval "tr/$maccaps/$maccaps/";
    $num_maccaps > $num_wincaps  ?  return 'm'  :  return 'w'
};

# If not, maybe it is in all capitals;
# count how many times there are "A, E, I, O"
for $enc (qw/k w /) {
    $num_AEIO{$enc} = eval "tr/$AEIO{$enc}/$AEIO{$enc}/";
    $num_AEIO{$enc} > $num_AEIO_max  &&  do {
        $num_AEIO_max = $num_AEIO{$enc}; $enc_AEIO = $enc
        }
    };

# suppose we found them (relative number more than $freqthr)
$num_AEIO_max > $thrhold * $length  &&  do {
    # Makes no sense to look for possible confusion with smalls.
    # It only remains to distinguish between 'a' and 'm'
    $enc_AEIO ne 'a'  &&  return $enc_AEIO;
    # to distinguish, first try smalls
    my $altsmalls = join '', map(chr, 160..175), map(chr, 224..239);
    my $macsmalls = join '', map(chr, 224..255);
    my $num_altsmalls = eval "tr/$altsmalls/$altsmalls/";
    my $num_macsmalls = eval "tr/$macsmalls/$macsmalls/";
    $num_altsmalls > $num_macsmalls  &&  return 'a';
    $num_macsmalls > $num_altsmalls  &&  return 'm';
    # if this doesn't work, try "YO", Ukrainian & Belorussian caps
    my $altmorecaps = join '', map(chr, qw/240 242 244 246 248 250/);
    my $macmorecaps = join '', map(chr, qw/221 162 184 167 186 216/);
    my $num_altmorecaps = eval "tr/$altmorecaps/$altmorecaps/";
    my $num_macmorecaps = eval "tr/$macmorecaps/$macmorecaps/";
    # if even this fails, return 'a' and hope they won't notice ;-)
    $num_macmorecaps > $num_altmorecaps  ?  return 'm'  :  return 'a'
};

# if all the above fails, we don't know what it is; return nothing
return 0;

}


##################################################################
sub usage { die
"Usage: geo-conv <xy> converts from x into y
       geo-conv <y> automatically converts into y
x = [kw], y = [kwl] (BPG/GOV/LATIN)\n"
}

##################################################################
__END__
GOV     BPG       LATIN      # Latin is given in GOV standard order
192      192      a          
193      193      b
194      194      g
195      195      d
196      196      e
197      197      v
198      198      z
199      225      E # e merve (e 8th) "old" geogian letters are in the "end" in BPG and UNICODE
200      199      T
201      200      i
202      201      k
203      202      l
204      203      m
205      204      n
206      226      I  # iota
207      205      o
208      206      p
209      207      J
210      208      r
211      209      s
212      210      t
213      227      V   # vi
214      211      u
215      212      f
216      213      q
217      214      R
218      215      y       
219      216      S
220      217      C       
221      218      c       
222      219      Z
223      220      w
224      221      W
225      222      x       
226      228      X   # khe
227      223      j
228      224      h    
229      229      H    # hoi    last letter in georgian alfabet (gov standard)
230      224      a    # from here start alternate letters currently presented only in BPG standar
